#!/bin/python


import sys
import os
import socket
import random
import resource
from ctypes import (CDLL, CFUNCTYPE, c_char, c_char_p, c_int, c_void_p,
                    c_size_t, c_uint32, get_errno, py_object, sizeof)

_NETLINK_ROUTE = 0

LIBNL = CDLL('libnl-3.so.200', use_errno=True)
LIBNL_ROUTE = CDLL('libnl-route-3.so.200', use_errno=True)

LIBNLv1 = None

def _init_libnlv1():
    global LIBNLv1
    global _nlv1_handle_alloc
    global _nlv1_handle_destroy
    global _nlv1_connect

    if LIBNLv1 is not None:
        return
    LIBNLv1 = CDLL('libnl.so.1', use_errno=True)

    _nlv1_handle_alloc = CFUNCTYPE(c_void_p)(('nl_handle_alloc', LIBNLv1))
    _nlv1_handle_destroy = CFUNCTYPE(None, c_void_p)(('nl_handle_destroy', LIBNLv1))
    _nlv1_connect = CFUNCTYPE(c_int, c_void_p, c_int)(('nl_connect', LIBNLv1))

_nl_socket_alloc = CFUNCTYPE(c_void_p)(('nl_socket_alloc', LIBNL))
_nl_socket_free = CFUNCTYPE(None, c_void_p)(('nl_socket_free', LIBNL))
_nl_connect = CFUNCTYPE(c_int, c_void_p, c_int)(('nl_connect', LIBNL))
_nl_geterror = CFUNCTYPE(c_char_p)(('nl_geterror', LIBNL))
_nl_socket_set_local_port = CFUNCTYPE(None, c_void_p, c_uint32)(('nl_socket_set_local_port', LIBNL))
_nl_socket_get_local_port = CFUNCTYPE(c_uint32, c_void_p)(('nl_socket_get_local_port', LIBNL))
_nl_socket_get_fd = CFUNCTYPE(c_int, c_void_p)(('nl_socket_get_fd', LIBNL))




def get_process_id():
    return os.getpid()

# libnl autogenerate at most 1024 ports. Return the list
# of the ports that it would generate (depends on the process id)
# See generate_local_port() in socket.c:
#   https://github.com/thom311/libnl/blob/fa380b409a02fe17b2d5cfc9074a8913523dbb68/lib/socket.c#L65
def get_port_numbers(process_id=None):
    if process_id is None:
        process_id = get_process_id()
    process_id = process_id & 0x3FFFFF;
    return [ (process_id | (x << 22)) for x in range (0, 1024) ]


def random_shuffle(num_elements, lst):
    lst = list(lst)
    random.shuffle(lst)
    i = 0;
    while i < num_elements:
        yield lst[i]
        i = i + 1


def bind_ports(num_sockets=None, pids=None):
    if pids is None:
        pids = get_port_numbers()
    if num_sockets is not None:
        pids = random_shuffle(num_sockets, pids)
    sockets = []
    for pid in pids:
        s = socket.socket(socket.AF_NETLINK,
                          socket.SOCK_RAW,
                          socket.NETLINK_ROUTE)
        s.bind( (pid, 0) )
        sockets.append(s)
    return sockets

def close_sockets(sockets):
    for s in sockets:
        s.close()

def nl_get_socket(do_connect=True, pid=None, libnlv1=False):

    if libnlv1:
        sk = _nlv1_handle_alloc()
    else:
        sk = _nl_socket_alloc()

    if sk is None:
        raise IOError(get_errno(), 'Failed to allocate netlink handle')

    if pid is not None:
        if libnlv1:
            raise Exception("Not implemented for libnl1")
        _nl_socket_set_local_port(sk, pid)

    if do_connect:
        if libnlv1:
            err = _nlv1_connect(sk, _NETLINK_ROUTE)
        else:
            err = _nl_connect(sk, _NETLINK_ROUTE)
            if err == 0:
                fd = _nl_socket_get_fd (sk)
                if fd <= 0:
                    raise Exception("failure while connecting netlink socket: fd is negative")
        if err != 0:
            if libnlv1:
                _nlv1_handle_destroy(sk)
                raise Exception("failed to connect libnl1 handle")
            else:
                _nl_socket_free(sk)
                raise IOError(-err, _nl_geterror())
    return sk

def free_nl_sockets(sks):
    for sk in sks:
        _nl_socket_free(sk)


#######################################################
# START
#######################################################

def TEST_a():
    # libnl would only try once to bind to a port. If the port was already in use
    # by another application, it would fail (without retrying another port). Now,
    # libnl will retry with another port.
    # see https://github.com/thom311/libnl/commit/4dd5fdd0af2c0b7ffe1dbc49313f263dbb2e906f
    # Test this by allocating 1000 conflicting ports outside of libnl's knowledge.
    # There is a high chance (1000/1024) that libnl's generated port will
    # fail with EADDRINUSE. New libnl will retry another port.

    print("TEST (a)...")

    sockets = bind_ports(1000)

    sk = nl_get_socket()

    close_sockets(sockets)
    _nl_socket_free (sk)

    print("...done")


def TEST_b():
    # old libnl would always try to generate a local port upon
    # creating the socket. As we at most try to generate 1024
    # ports, you could at most create 1024 socket instances.
    # Now, we don't generate the local port before actually needing
    # it. Especially, not during creation of the socket, and you can create
    # more sockets:
    # see https://github.com/thom311/libnl/commit/0271578987088210d7d2d68addbd5e8fe27d4383
    print("TEST (b)...")

    sks = []

    for i in range(0, 1500):
        sk = nl_get_socket(do_connect=False)
        sks.append(sk)

    free_nl_sockets(sks)
    print("...done")

def TEST_c():
    # libnl3 would generate the ports in the same order as libnl1
    # (and both libraries would not retry). If you happen to use both libraries
    # in your application (for example indirectly loading a library that uses libnl1),
    # then the library which tries to connect second would fail with EADDRINUSE --
    # without retrying.
    # Now, libnl generates the port in a different order then libnl1. Thus it leaves the
    # first 8 ports the libnl1 preferably uses.
    # see https://github.com/thom311/libnl/commit/1f734a8f892abcd3f81637df4a089155aca1b66a
    # Test this by connecting first a libnl3 socket, and then a libnl1 socket.

    print("TEST (c)...")

    _init_libnlv1()

    sk = nl_get_socket()

    skv1 = nl_get_socket(libnlv1=True)

    _nl_socket_free(sk)
    _nlv1_handle_destroy(skv1)

    print("...done")

def TEST_d():
    # libnl only tried to generate at most 1024 local ports. After that number was
    # exhausted, it would not try again, and you could not connect more sockets
    # (unless selecting the local port manually and thus working around the failure
    # autogenerating the port).
    # Now, if the ports are exhausted, libnl3 leaves the port unspecified and lets
    # kernel choose. Thus you can connect more then 1024 ports.
    # see https://github.com/thom311/libnl/commit/96e1e5bdc2e803700055395cc3c428fa2525d1ca

    print("TEST (d)...")

    sks = []
    for i in range(0, 1500):
        sk = nl_get_socket()
        #print("connected socket #%s: fd=%s, pid=%s" % (i, _nl_socket_get_fd(sk), _nl_socket_get_local_port(sk)))
        sks.append(sk)

    free_nl_sockets(sks)
    print("...done")



rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (2048 if 2048 < rlimit[1] else rlimit[1], rlimit[1]))
print("ulimit(NOFILE) = %s" % (str(resource.getrlimit(resource.RLIMIT_NOFILE))))

print("Test: PID=%s" % (get_process_id()))


if len(sys.argv) == 1:
    argv = ["a"]
else:
    argv = list(sys.argv[1:])

for arg in argv:
    locals()["TEST_" + arg]()


